from __future__ import annotations

"""Bonus-specific deterministic hooks used by sim and presentation paths."""

from .pickup_fx import emit_bonus_pickup_effects

__all__ = ["emit_bonus_pickup_effects"]
