from __future__ import annotations

from importlib.metadata import version

__version__ = version("crimsonland")

__all__ = [
    "atlas",
    "audio_router",
    "bonuses",
    "creatures",
    "gameplay",
    "effects",
    "effects_atlas",
    "modes",
    "perks",
    "persistence",
    "quests",
    "replay",
    "render",
    "sim",
    "ui",
    "views",
    "weapons",
]
