from __future__ import annotations

"""Presentation-phase deterministic hook registries."""

from .projectile_decals import ProjectileDecalCtx, run_projectile_decal_hooks

__all__ = ["ProjectileDecalCtx", "run_projectile_decal_hooks"]

